#include <fstream.h>
// #include <iomanip.h>
#include <strstream.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>

static
char
pnm_getc(istream &istr)		// 1st non-comment, non-whitespace, 0 on failure
{
	char c;
	while (istr.get(c) && (c == ' ' || c == '\t' || c == '\n'));
	if (!istr) return 0;
	if (c == '#') {		// skip comments
		while (istr.get(c) && c != '\n');
		if (!istr) return 0;
		while (istr.get(c) && (c == ' ' || c == '\t' || c == '\n'));
		if (!istr) return 0;
	}
	return c;
}

static
int
pnm_getu(istream &istr)		// get unsigned value, -1 on failure
{
	char c=pnm_getc(istr);	// skips comments and whitespace
	if(!c || !isdigit(c)) return -1;
	unsigned value=0;
	while (c && isdigit(c)) {
		value*=10;
		value+=(c-'0');
		istr.get(c);
	}
	// discards single trailing white space
	return value;
}

static
unsigned short
pnm_geti(istream &istr,unsigned short wordlength)	// get image data value
{
	unsigned char c[2];
	istr.read(c,wordlength);
	unsigned short value;
	// PNM format is little-endian ...
	if (wordlength == 2)
		value=(c[1]<<8)+c[0];
	else
		value=c[0];
	return value;
}

int
main(int argc, char *argv[])
{
	int verbose = 1;
	int err = 0;

	if (argc != 3) {
		cerr << "Must specify two filenames as input" << endl;
		err=1;
	}

	if (err) {
		cerr << "Usage: " << argv[0]
		     << " filea fileb\n"
		     << flush;
		return 1;
	}

	char *filenamea = argv[1];
	char *filenameb = argv[2];

	ifstream *ina = new ifstream(filenamea);
	if (!ina->rdbuf()->is_open()) {
		cerr << "Can't open " << filenamea << "" << endl;
		return 1;
	}
	ifstream *inb = new ifstream(filenameb);
	if (!inb->rdbuf()->is_open()) {
		cerr << "Can't open " << filenameb << "" << endl;
		return 1;
	}

	// Read PNM headers ...

	char magica=pnm_getc(*ina);
	char formata;
	ina->get(formata);

	if (ina->fail() || magica != 'P') {
		cerr << "Not a pnm file " << filenamea << "" << endl;
		return 1;
	}

	if (formata != '5') {
		cerr << "Unsupported pnm format P" << formata
		     << " in " << filenamea
		     << "" << endl;
		return 1;
	}

	int widtha = pnm_getu(*ina);
	int heighta = pnm_getu(*ina);
	int maxvala = pnm_getu(*ina);
	unsigned wordlengtha=(maxvala <= 255) ? 1 : 2;

	if (!*ina || widtha == -1 || heighta == -1 || maxvala == -1) {
		cerr << "Corrupt pnm header " << filenamea << "" << endl;
		return 1;
	}

	char magicb=pnm_getc(*inb);
	char formatb;
	inb->get(formatb);

	if (inb->fail() || magicb != 'P') {
		cerr << "Not a pnm file " << filenameb << "" << endl;
		return 1;
	}

	if (formatb != '5') {
		cerr << "Unsupported pnm format P" << formatb
		     << " in " << filenameb
		     << "" << endl;
		return 1;
	}

	int widthb = pnm_getu(*inb);
	int heightb = pnm_getu(*inb);
	int maxvalb = pnm_getu(*inb);
	unsigned wordlengthb=(maxvalb <= 255) ? 1 : 2;

	if (!*inb || widthb == -1 || heightb == -1 || maxvalb == -1) {
		cerr << "Corrupt pnm header " << filenameb << "" << endl;
		return 1;
	}

	if (verbose) {
		cerr << "File " << filenamea << "" << endl;
		cerr << "\twidth " << widtha << "" << endl;
		cerr << "\theight " << heighta << "" << endl;
		cerr << "\tmaxval " << maxvala << "" << endl;
		cerr << "\tbytes per pixel " << wordlengtha << "" << endl;
		cerr << "File " << filenameb << "" << endl;
		cerr << "\twidth " << widthb << "" << endl;
		cerr << "\theight " << heightb << "" << endl;
		cerr << "\tmaxval " << maxvalb << "" << endl;
		cerr << "\tbytes per pixel " << wordlengthb << "" << endl;
	}

	if (formata != formatb
	 || widtha != widthb
	 || heighta != heightb
	 ) {
		cerr << "PNM files different formats or sizes :(" << endl;
		return 1;
	}

	// Last whitespace discarded ... at start of data

	int width  = widtha;
	int height = heighta;

	// Process image data ...

	double sum1squared=0;
	double sumdeltasquared=0;
	unsigned y;
	for (y=0; y<height; ++y) {
		unsigned x;
		for (x=0; x<width; ++x) {
			unsigned short value1=pnm_geti(*ina,wordlengtha);
			unsigned short value2=pnm_geti(*inb,wordlengthb);

			sum1squared+=(double)value1*(double)value1;
			sumdeltasquared+=((double)value2-(double)value1)
					*((double)value2-(double)value1);
		}
	}

	double nmse=sumdeltasquared/sum1squared;
	double rmse=sqrt(sumdeltasquared/((double)width*(double)height));

	if (ina->fail()) {
		cerr << "Read error on " << filenamea << "" << endl;
		return 1;
	}
	if (inb->fail()) {
		cerr << "Read error on " << filenameb << "" << endl;
		return 1;
	}

	if (verbose)
		cerr << "sumdeltasquared = " << sumdeltasquared
		     << " sum1squared = " << sum1squared
		     << endl;

	cerr << "nmse = " << nmse
	     << " rmse = " << rmse
	     << endl;

	return 0;
}

